from commons.core.sql.TablePathAdaptator import TablePathAdaptator
from commons.core.sql.TableSeqAdaptator import TableSeqAdaptator
from PASTEC.utils.AbstractAnalyseAgent import AbstractAnalyseAgent
from PASTEC.DemandAndResult import DemandAndResult
from PASTEC.BlasterMatch import BlasterMatch

## Class to create a new BLRAgent. A BLRAgent is an Agent that retrieve informations about blast against a TE bank (i.e. Repbase)
# 
class Abstract_TE_BLRx_txAgent(AbstractAnalyseAgent):

    def __init__(self, env):
        AbstractAnalyseAgent.__init__(self, env)
        self._iTableSeqAdaptator = TableSeqAdaptator(self._db, self.getRepbaseTableName())
    
    def getRepbaseTableName(self):
        return self._environment.getProjectName() + self.SUFFIX_TABLE_NAME_LIST[1]
    
    def getListOfFeaturesToAnalyse(self, consensusName):
        iTablePathAdaptator = TablePathAdaptator(self._db, self._tableName)
        return iTablePathAdaptator.getPathListFromQuery(consensusName)
    
    def getSubjectSeqLength(self, subjectName):
        return self._iTableSeqAdaptator.getSeqLengthFromAccession(subjectName)
    
    ## Analyze a Demand
    #
    # @param demand a Demand instance
    # @param lPaths list of path instances
    # @return a DemandAndResult instance with a BlasterMatch list
    #
    def analyse(self, demand, lPaths):
        lSortedPaths = self._sortPathsListBySubjectNameStartEnd(lPaths)
        lBlasterMatchs = self._createBlasterMatchsList(lSortedPaths)
        return DemandAndResult(demand, lBlasterMatchs)
    
    def _sortPathsListBySubjectNameStartEnd(self, lPaths):
        return sorted(lPaths, key=lambda k: (k.getSubjectName(), k.getSubjectStart(), k.getSubjectEnd()))
    
    def _createBlasterMatchsList(self, lPaths):
        lBM = []
        lPathsOnDirectStrand = []
        lPathsOnReverseStrand = []
        for iPath in lPaths:
            if iPath.isSubjectOnDirectStrand():
                lPathsOnDirectStrand.append(iPath)
            else:
                lPathsOnReverseStrand.append(iPath)
        if lPathsOnDirectStrand != []:
            self._createBlasterMatchListForOneStrand(lPathsOnDirectStrand, lBM, "+")
        if lPathsOnReverseStrand != []:
#            lPathsOnReverseStrands = sorted(lPathsOnReverseStrand, key=lambda k: (k.getSubjectName(), k.getSubjectEnd(), k.getSubjectStart()))
            self._createBlasterMatchListForOneStrand(lPathsOnReverseStrand, lBM, "-")
        return lBM

    def _createBlasterMatchListForOneStrand(self, lPaths, lBM, strand):
        coverageLength = 0
        minCoord, maxCoord = self._initializeCoords(lPaths[0])
        previousPath = lPaths[0]
        for iPath in lPaths[1:]:
            if previousPath.getSubjectName() == iPath.getSubjectName():
                if iPath.isSubjectOverlapping(previousPath):
#                    maxCoord = max(previousPath.getSubjectEnd(), iPath.getSubjectEnd(), maxCoord)
                    maxCoord = max(iPath.getSubjectStart(), iPath.getSubjectEnd(), maxCoord)
                else:
                    coverageLength += maxCoord - minCoord + 1
                    minCoord, maxCoord = self._initializeCoords(iPath)
            else:
                self._computeCoverageAndCreateBM(lBM, coverageLength, minCoord, maxCoord, previousPath, strand)
                coverageLength = 0
                minCoord, maxCoord = self._initializeCoords(iPath)
            previousPath = iPath
        self._computeCoverageAndCreateBM(lBM, coverageLength, minCoord, maxCoord, previousPath, strand)
        
    def _computeCoverageAndCreateBM(self, lBM, coverageLength, minCoord, maxCoord, previousPath, strand):
        coverageLength += maxCoord - minCoord + 1
        coverage = coverageLength * 100 / float(self.getSubjectSeqLength(previousPath.getSubjectName()))
        if coverage >= self._environment.getMinBlastIndex(): 
            #TODO: investigate : why is the identity not passed to BlasterMatch ? we end up with identity =0
            iBlasterMatch = BlasterMatch(previousPath.getSubjectName(), coverage, strand)
            lBM.append(iBlasterMatch)

    def _initializeCoords(self, iPath):
        minCoord = min(iPath.getSubjectStart(), iPath.getSubjectEnd())
        maxCoord = max(iPath.getSubjectStart(), iPath.getSubjectEnd())
        return minCoord, maxCoord
    
#    #TODO: moy des identites ?
#    def _computeIndex(self, iPath, coverage):
#        blastIdWeight = self._searchCorrespondingBlastIdWeight(iPath.getIdentity())
#        return coverage * 100 * blastIdWeight
#    
#    def _searchCorrespondingBlastIdWeight(self, identity):
#        blastIdWeight = 0
#        blastIdWeightTuples = self._environment.getBlastIdWeight()
#        for blastIdWeightTuple in blastIdWeightTuples:
#            if identity > blastIdWeightTuple[0] and identity <= blastIdWeightTuple[1]:
#                blastIdWeight = blastIdWeightTuple[2]
#        return blastIdWeight