from commons.core.sql.TablePathAdaptator import TablePathAdaptator
from commons.core.sql.TableSeqAdaptator import TableSeqAdaptator
from PASTEC.utils.AbstractAnalyseAgent import AbstractAnalyseAgent
from PASTEC.MatchingSubject import MatchingSubject
from PASTEC.HelitronExtremity import HelitronExtremity
from PASTEC.DemandAndResult import DemandAndResult
import re

## This agent search in the database the best blastn match and the helitron extremities
#
class TE_BLRn_Agent(AbstractAnalyseAgent):
    
    SUFFIX_TABLE_NAME_LIST = ["_TE_BLRn_path", "_RepbaseNT_seq"] 
    
    def __init__(self, env):
        AbstractAnalyseAgent.__init__(self, env)
        self._iTableSeqAdaptator = TableSeqAdaptator(self._db, self.getRepbaseTableName())
    
    def _setAgentDecisionRules(self):
        """sets the current agen'ts decision rules based on info from configuration file"""
        decisionRulesData = self._environment.getAgentDecisionRules(self.__class__.__name__)
        self.MAX_ERROR_WINDOW = decisionRulesData["MAX_ERROR_WINDOW"]
    
    def getRepbaseTableName(self):
        return self._environment.getProjectName() + self.SUFFIX_TABLE_NAME_LIST[1]
    
    def getSubjectSeqLength(self, subjectName):
        return self._iTableSeqAdaptator.getSeqLengthFromAccession(subjectName)

    def getListOfFeaturesToAnalyse(self, consensusName):
        tpA = TablePathAdaptator(self._db, self._tableName)
        lPath = tpA.getPathListFromQuery(consensusName)
        lMatchingSubject = []
        for path in lPath:
            length = self.getSubjectSeqLength(path.getSubjectName())
            matchingSubject = MatchingSubject(path, length)
            lMatchingSubject.append(matchingSubject)
        return lMatchingSubject
            
    def analyse(self, demand, lMatchingSubject):
        consensusLength = self.askForConsensusLength(demand.consensusName)
        lMatchingSubjectSorted = self._sortLPathBySubjectNameAndQueryStartAndByQueryEnd(lMatchingSubject)
        lHelitronExtremities = self._searchHelitronsExtremities(lMatchingSubjectSorted, consensusLength)
        lHelitronExtremitiesSorted = self._sortHelitronExtremities(lHelitronExtremities)
        result = DemandAndResult(demand, lHelitronExtremitiesSorted)
        return result
    
    def _searchHelitronsExtremities(self, lMatchingSubject, consensusLength):
        lHelitronExtremities = []
        for matchingSubject in lMatchingSubject:
            if (re.search("Helitron", matchingSubject.range_subject.seqname)) \
            and self._isMatchingSubjectAtExtremities(matchingSubject, consensusLength) and matchingSubject.e_value < self._environment.getMaxBlastnEvalue():
                iHelitronExtremity = HelitronExtremity()
                iHelitronExtremity.setName(matchingSubject.range_subject.seqname)
                iHelitronExtremity.setEValue(matchingSubject.e_value)
                iHelitronExtremity.setStart(matchingSubject.range_query.start)
                iHelitronExtremity.setEnd(matchingSubject.range_query.end)
                lHelitronExtremities.append(iHelitronExtremity)
        return lHelitronExtremities

    def _isMatchingSubjectAtExtremities(self, matchingSubject, consensusLength):
        return self._IsStartOnQueryAndSubjectInPath(matchingSubject) or self._IsEndOnQueryAndSubjectInPath(consensusLength, matchingSubject)

    def _IsQueryStart(self, matchingSubject):
        return matchingSubject.range_query.start < self.MAX_ERROR_WINDOW

    def _IsSubjectStartDirect(self, matchingSubject):
        return matchingSubject.range_subject.start < self.MAX_ERROR_WINDOW

    def _IsSubjectStartReverse(self, matchingSubject):
        return matchingSubject.range_subject.start > (matchingSubject._length - self.MAX_ERROR_WINDOW)

    def _IsQueryEnd(self, consensusLength, matchingSubject):
        return matchingSubject.range_query.end > (consensusLength - self.MAX_ERROR_WINDOW) 

    def _isSubjectEndDirect(self, matchingSubject):
        return matchingSubject.range_subject.end > (matchingSubject._length - self.MAX_ERROR_WINDOW)

    def _isSubjectEndReverse(self, matchingSubject):
        return matchingSubject.range_subject.end < self.MAX_ERROR_WINDOW

    def _IsStartOnQueryAndSubjectInPath(self, matchingSubject):
        return (self._IsQueryStart(matchingSubject)) and ((self._IsSubjectStartDirect(matchingSubject)) or (self._IsSubjectStartReverse(matchingSubject)))

    def _IsEndOnQueryAndSubjectInPath(self, consensusLength, matchingSubject):
        return (self._IsQueryEnd(consensusLength, matchingSubject)) and ((self._isSubjectEndDirect(matchingSubject)) or (self._isSubjectEndReverse(matchingSubject)))

    def _sortLPathBySubjectNameAndQueryStartAndByQueryEnd(self, lPath):
        return sorted( lPath, key=lambda iPath: ( iPath.range_subject.seqname, iPath.range_query.start, iPath.range_query.end ))
    
    def _sortHelitronExtremities(self, lHelitronExtremities):
        return sorted( lHelitronExtremities, key=lambda iHelitronExtremity: ( iHelitronExtremity.getStart(), iHelitronExtremity.getEnd() ))
    