#!/usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import math
import os
from commons.core.LoggerFactory import LoggerFactory
from commons.core.utils.RepetOptionParser import RepetOptionParser
from commons.core.utils.FileUtils import FileUtils
from commons.core.parsing.VarscanFile import VarscanFile
from commons.core.seq.Bioseq import Bioseq

LOG_DEPTH = "core.parsing"

##Converts varscan 2.2.8 file to VCF file 4.1
#
class VarscanToVCF(object):
    
    def __init__(self, varscanFileName = "", vcfFileName = "", doClean = False, verbosity = 0):
        self._varscanFileName = varscanFileName
        self.setVcfFileName(vcfFileName)
        self._doClean = doClean
        self._verbosity = verbosity
        
        self._vcfRevision = "VCFv4.1"
        self._vcfHeader = "#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO"
        
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self._verbosity)
        
    def setAttributesFromCmdLine(self):
        description = "Convert Varscan file (v2.2.8) to VCF file (v4.1)."
        epilog = "Example:\npython VarscanToVCF.py -i varscanFileName -v 2\n"
        parser = RepetOptionParser(description = description, epilog = epilog)
        parser.add_option("-i", "--Varscan",    dest = "varscanFileName", action = "store",       type = "string", help = "input Varscan file name [compulsory] [format: varscan2.2.8]", default = "")
        parser.add_option("-o", "--vcfFileName",dest = "vcfFileName",     action = "store",       type = "string", help = "vcfFileName file name [default: <input>.vcf]", default = "")
        parser.add_option("-c", "--clean",      dest = "doClean",         action = "store_true",                   help = "clean temporary files [optional] [default: False]", default = False)
        parser.add_option("-v", "--verbosity",  dest = "verbosity",       action = "store",       type = "int",    help = "verbosity [optional] [default: 1]", default = 1)
        options = parser.parse_args()[0]
        self._setAttributesFromOptions(options)
        
    def _setAttributesFromOptions(self, options):
        self.setVarscanFileName(options.varscanFileName)
        self.setVcfFileName(options.vcfFileName)
        self.setDoClean(options.doClean)
        self.setVerbosity(options.verbosity)

    def setVarscanFileName(self, varscanFileName):
        self._varscanFileName = varscanFileName
        
    def setVcfFileName(self, vcfFileName):
        if vcfFileName == "":
            self._vcfFileName = "%s.vcf" % self._varscanFileName
        else:
            self._vcfFileName = vcfFileName
        
    def setDoClean(self, doClean):
        self._doClean = doClean
        
    def setVerbosity(self, verbosity):
        self._verbosity = verbosity
        
    def _checkOptions(self):
        if self._varscanFileName == "":
            self._logAndRaise("ERROR: Missing input file name")
        else:
            if not FileUtils.isRessourceExists(self._varscanFileName):
                self._logAndRaise("ERROR: Input Varscan file '%s' does not exist!" % self._varscanFileName)
            
    def _logAndRaise(self, errorMsg):
        self._log.error(errorMsg)
        raise Exception(errorMsg)

    def _convertVarscanLineToVCFRecord(self, lVarscanLine, lLineNumber):
        vcfLine = ""
        iVarscanFile = VarscanFile()
        iVarscanFile.setTypeOfVarscanFile("Varscan_2_2_8")
        
        lVarscanHit = []
        for varLine in lVarscanLine:
            lVarscanHit.append(iVarscanFile.createVarscanObjectFromLine(varLine, lLineNumber.pop(0)))
        
        if len(lVarscanHit) == 1:
            iVarscanHit = lVarscanHit[0]
            Chrom = iVarscanHit.getChrom()
            Pos = int(iVarscanHit.getPosition())
            ID = "."
            Ref = iVarscanHit.getRef()
            Alt = iVarscanHit.getVar()
            if float(iVarscanHit.getPValue()) == 0.98:
                Qual = "."
            elif float(iVarscanHit.getPValue()) == 0:
                Qual = 10000
            else:
                Qual = -10*math.log10(float(iVarscanHit.getPValue()))
            Filter = "."
            AF = float(iVarscanHit.getVarFreq()[:-1])/100
            DP = int(iVarscanHit.getReadsRef()) + int(iVarscanHit.getReadsVar())
            RBQ = iVarscanHit.getQualRef()
            ABQ = iVarscanHit.getQualVar()
            Info = ";".join(["AF=%.4f" %AF,"DP=%d" %DP,"RBQ=%s" %RBQ, "ABQ=%s" %ABQ])
            
            if iVarscanHit.getCns() in ["R", "Y", "M", "K", "W", "S"]:
                allel = Bioseq().getATGCNFromIUPACandATGCN(iVarscanHit.getCns(), Ref)
            else:
                allel = iVarscanHit.getCns()
                
            if allel != Alt and len(Alt) < 2:
                self._log.warning("'VarAllele' attribute of Varscan file was not correct for position '%s'. Correcting using '%s' instead of '%s'." % (Pos, allel, Alt))
                Alt = allel
            
            if Qual == ".":
                vcfLine = "%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n" % (Chrom, Pos, ID, Ref, Alt, Qual, Filter, Info)
            elif Qual == 10000:
                vcfLine = "%s\t%s\t%s\t%s\t%s\t%d\t%s\t%s\n" % (Chrom, Pos, ID, Ref, Alt, Qual, Filter, Info)
            else:
                vcfLine = "%s\t%s\t%s\t%s\t%s\t%.9f\t%s\t%s\n" % (Chrom, Pos, ID, Ref, Alt, Qual, Filter, Info)
        elif len(lVarscanHit) > 1:
            iVarscanHit1 = lVarscanHit[0]
            iVarscanHit2 = lVarscanHit[1]
            
            Chrom = iVarscanHit1.getChrom()
            Pos = int(iVarscanHit1.getPosition())
            ID = "."
            Ref = iVarscanHit1.getRef()
            
            Alt1 = iVarscanHit1.getVar()
            if iVarscanHit1.getCns() in ["R", "Y", "M", "K", "W", "S"]:
                allel1 = Bioseq().getATGCNFromIUPACandATGCN(iVarscanHit1.getCns(), Ref)
            else:
                allel1 = iVarscanHit1.getCns()
            if allel1 != Alt1 and len(Alt1) < 2:
                self._log.warning("'VarAllele' attribute of Varscan file was not correct for position '%s'. Correcting using '%s' instead of '%s'." % (Pos, allel1, Alt1))
                Alt1 = allel1
                
            if float(iVarscanHit1.getPValue()) == 0.98 :
                Qual = "."
            elif float(iVarscanHit1.getPValue()) == 0:
                Qual = 10000
            else:
                Qual = -10*math.log10(float(iVarscanHit1.getPValue()))
            Filter = "."
            AF1 = float(int(iVarscanHit1.getReadsVar()))/(int(iVarscanHit1.getReadsVar()) + int(iVarscanHit2.getReadsVar()) + int(iVarscanHit1.getReadsRef()))
            RBQ1 = iVarscanHit1.getQualRef()
            ABQ1 = iVarscanHit1.getQualVar()
            
            Alt2 = iVarscanHit2.getVar()
            if iVarscanHit2.getCns() in ["R", "Y", "M", "K", "W", "S"]:
                allel2 = Bioseq().getATGCNFromIUPACandATGCN(iVarscanHit2.getCns(), Ref)
            else:
                allel2 = iVarscanHit2.getCns()
            if allel2 != Alt2 and len(Alt2) < 2:
                self._log.warning("'VarAllele' attribute of Varscan file was not correct for position '%s'. Correcting using '%s' instead of '%s'." % (Pos, allel2, Alt2))
                Alt2 = allel2
                
            AF2 = float(int(iVarscanHit2.getReadsVar()))/(int(iVarscanHit1.getReadsVar()) + int(iVarscanHit2.getReadsVar()) + int(iVarscanHit1.getReadsRef()))
            DP = int(iVarscanHit1.getReadsRef()) + int(iVarscanHit1.getReadsVar()) + int(iVarscanHit2.getReadsVar())
            ABQ2 = iVarscanHit2.getQualVar()
            
            Info = ";".join(["AF=%.4f,%.4f" % (AF1, AF2), "DP=%d" % DP, "RBQ=%s" % RBQ1, "ABQ=%s,%s" % (ABQ1, ABQ2)])
            Alt = "%s,%s" % (Alt1, Alt2)
            
            if Qual == ".":
                vcfLine = "%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n" % (Chrom, Pos, ID, Ref, Alt, Qual, Filter, Info)
            elif Qual == 10000:
                vcfLine = "%s\t%s\t%s\t%s\t%s\t%d\t%s\t%s\n" % (Chrom, Pos, ID, Ref, Alt, Qual, Filter, Info)
            else:
                vcfLine = "%s\t%s\t%s\t%s\t%s\t%.9f\t%s\t%s\n" % (Chrom, Pos, ID, Ref, Alt, Qual, Filter, Info)
            
        return vcfLine
    
    def _hasVarscanHeader(self):
        hasVarscanHeader = False
        varscanHeader = "Chrom\tPosition\tRef\tCons\tReads1\tReads2\tVarFreq\tStrands1\tStrands2\tQual1\tQual2\tPvalue\tMapQual1\tMapQual2\tReads1Plus\tReads1Minus\tReads2Plus\tReads2Minus\tVarAllele"
        with open(self._varscanFileName, "r") as f:
            l = f.readline().rstrip().lstrip("#")
            if l == varscanHeader:
                hasVarscanHeader = True
        return hasVarscanHeader

    def run(self):
        LoggerFactory.setLevel(self._log, self._verbosity)
        self._checkOptions()
        self._log.info("START Varscan To VCF")
        self._log.warning("Only SNPs will be converted")
        self._log.debug("Input file name: %s" % self._varscanFileName)
        
        with open(self._vcfFileName, "w") as fVCF:
            fVCF.write("##fileformat=%s\n" % self._vcfRevision)
            fVCF.write("##qual=10000 if varscan p-value=0\n")
            fVCF.write("##program: VarscanToVCF.py input_file=%s\n" % os.path.basename(self._varscanFileName))
            fVCF.write("%s\n" % self._vcfHeader)
            
            with open(self._varscanFileName, "r") as fVarscan:
                lineNumber1 = 2
                lineNumber2 = 3
                if self._hasVarscanHeader():
                    fVarscan.readline()
                line1 = fVarscan.readline()
                line2 = fVarscan.readline()
                while line1:
                    string1 = "\t".join([line1.split()[0], line1.split()[1]])
                    if line2:
                        string2 = "\t".join([line2.split()[0], line2.split()[1]])
                    else:
                        string2 = ""
                    if string1 == string2:
                        lLine = [line1, line2]
                        vcfLine = self._convertVarscanLineToVCFRecord(lLine, [lineNumber1, lineNumber2])
                        line1 = fVarscan.readline()
                        line2 = fVarscan.readline()
                        lineNumber1 += 2
                        lineNumber2 += 2
                    else:
                        lLine = [line1]
                        vcfLine = self._convertVarscanLineToVCFRecord(lLine, [lineNumber1])
                        line1 = line2
                        line2 = fVarscan.readline()
                        lineNumber1 = lineNumber2
                        lineNumber2 += 1
                        
                    if "-" not in vcfLine.split()[4] and "+" not in vcfLine.split()[4]:
                        fVCF.write(vcfLine)
                                
        self._log.info("END Varscan To VCF")

if __name__ == "__main__":
    iLaunch = VarscanToVCF()
    iLaunch.setAttributesFromCmdLine()
    iLaunch.run()        