#!/usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import os
import glob
import subprocess
from commons.core.LoggerFactory import LoggerFactory
from commons.core.utils.RepetOptionParser import RepetOptionParser

LOG_DEPTH = "commons.core.launcher"

class LaunchBlat( object ):
    """
    Launch Blat (pairwise alignment).
    """
    
    def __init__(self, inFile = "", subjectFile = "", outFile = "", allByAll = False, clean = False, verbosity = 3):
        self._inFile = inFile
        self._sbjFile = subjectFile
        self._outFile = outFile
        self._allByAll = allByAll
        self._clean = clean
        self._verbose = verbosity
        
        self._prgName = "blat"
        
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self._verbose)
        
        #TODO: unused, remove these attributes?
#        self._wrpName = "%s.py" % type(self).__name__
#        self._wrpCmdLine = ""
#        self._prgCmdLine = ""
#        self._lFilesToKeep = []
#        self._lFilesToRemove = []
#        self._cmdLineGenericOptions = "hi:cv:"
#        self._cmdLineSpecificOptions = ""
#        self._formatInFile = "fasta"
#        self._prgParam = ""

        
    def setAttributesFromCmdLine(self):
        description = "Launch blat for pairwise alignment."
        epilog = "\nExample: launch without verbosity and keep temporary files.\n"
        epilog += "\t$ LaunchBlat.py -i file.fa -v 0\n"
        parser = RepetOptionParser(description = description, epilog = epilog)
        parser.add_option("-i", "--inFile",     dest = "fastaFileName",   action = "store",       type = "string", help = "input fasta file name [compulsory] [format: fasta]", default = "")
        parser.add_option("-s", "--subj",       dest = "subjectFileName", action = "store",       type = "string", help = "name of the subject file (database, format='fasta')", default = "")
        parser.add_option("-o", "--outFile",    dest = "outFilename",     action = "store",       type = "string", help = "name of the output file (format='align', default=inFile+'.align')",    default = "")
        parser.add_option("-A", "--allByAll",   dest = "allByAll",        action = "store_true",                   help = "same sequences (all-by-all)", default = False)
        parser.add_option("-c", "--clean",      dest = "doClean",         action = "store_true",                   help = "clean temporary files [optional] [default: False]", default = False)
        parser.add_option("-v", "--verbosity",  dest = "verbosity",       action = "store",       type = "int",    help = "verbosity [optional] [default: 3]", default = 3)
        options = parser.parse_args()[0]
        self._setAttributesFromOptions(options)
    
    def _setAttributesFromOptions(self, options):
        self.setInFileName(options.fastaFileName)
        self.setSubjectFile(options.subjectFileName)
        self.setOutFileName(options.outFilename)
        self.setAllByAll(options.allByAll)
        self.setClean(options.doClean)
        self.setVerbosity(options.verbosity)
            
    def setInFileName(self, arg):
        self._inFile = arg

    def getInFileName(self):
        return self._inFile    
            
    def setOutFileName(self, arg):
        self._outFile = arg

    def getOutFileName(self):
        return self._outFile    
            
    def setSubjectFile(self, arg):
        self._sbjFile = arg     
        
    def getSubjectFile(self):
        return self._sbjFile    
    
    def setAllByAll(self, arg):
        self._allByAll = True       
        
    def getAllByAll(self):
        return self._allByAll

    def setClean(self, arg):
        self._clean = arg       
        
    def setVerbosity(self, arg):
        self._verbose = arg       

    def _checkOptions(self):
        if self.getSubjectFile == "":
            self._logAndRaise("ERROR: missing subject file (-s)")
        if not os.path.exists(self.getSubjectFile()):
            self._logAndRaise("ERROR: subject file '%s' doesn't exist" % self.getSubjectFile())
        if self.getInFileName() == "":
            self._logAndRaise("ERROR: Missing input fasta file name")
        if self._outFile == "":
            self.setOutputFile("%s.align" % self._inFile)
            
    def _logAndRaise(self, errorMsg):
        self._log.error(errorMsg)
        raise Exception(errorMsg)       
        
    def convertBlastIntoAlign(self):
        """
        Convert a 'blast' file into the 'align' format.
        """
        cmd = "blast2align.py"
        cmd += " -i %s.blast" % self.getInFileName()
        cmd += " -o %s" % self.getOutFileName()
        process = subprocess.Popen(cmd, shell = True)
        self._log.debug("Running : %s" % cmd)
        process.communicate()
        if process.returncode != 0:
            self._logAndRaise("ERROR when launching '%s'" % cmd)
            
    def clean(self):
        """
        Remove the files in 'getListFilesToRemove()'.
        """
        lFilesToRemove = glob.glob("*.blast")
        for f in lFilesToRemove:
            if os.path.exists(f):
                os.remove(f)
        
    def run(self):
        LoggerFactory.setLevel(self._log, self._verbose)
        self._checkOptions()
        self._log.info("START LaunchBlat")
        self._log.debug("With input file: %s , subject file: %s, output file: %s, allByAll: %s" % (self.getInFileName(), self.getSubjectFile() , self.getOutFileName(), self.getAllByAll()))
        
        cmd = self._prgName
        cmd += " %s" % self.getSubjectFile()
        cmd += " %s" % self.getInFileName()
        cmd += " -out=blast8"
        cmd += " %s.blast" % self.getInFileName()
        process = subprocess.Popen(cmd, shell = True)
        self._log.debug("Running : %s" % cmd)
        process.communicate()
        if process.returncode != 0:
            self._logAndRaise("ERROR program Blat returned exit status '%i'" % process.returncode)
            
        self.convertBlastIntoAlign()
        
        if self._clean:
            self.clean()
            
        self._log.info("END LaunchBlat")
        
if __name__ == "__main__":
    iLaunch = LaunchBlat()
    iLaunch.setAttributesFromCmdLine()
    iLaunch.run()