#!/usr/bin/env python

from PASTEC.SuperAgent import SuperAgent
from PASTEC.Environment import Environment
from commons.pyRepetUnit.profilesDB.ProfilesDB2Map import ProfilesDB2Map
from commons.core.sql.DbFactory import DbFactory
from commons.core.utils.RepetOptionParser import RepetOptionParser
from commons.core.utils.FileUtils import FileUtils
import ConfigParser
import sys
import os

#TODO: check env var if launch as tool (without TEdenovo)
class LaunchPASTEC(object):
    
    def __init__(self, configFileName = "", decisionRulesFileName = "",  step = "12", inputFileName = "", verbose = 0, projectName = ""):
        self._configFileName = configFileName
        self._decisionRulesFileName = decisionRulesFileName
        self._projectName = projectName
        self._step = step
        self._inFileName = inputFileName
        self._verbose = verbose
        
    def setConfigFileName(self, configFileName):
        self._configFileName = configFileName

    def setDecisionRulesFileName(self, decisionRules):
        self._decisionRulesFileName = decisionRules    
        
    def setProjectName(self, projectName):
        self._projectName = projectName
        
    def setStep(self, step):
        self._step = step
        
    def setInputFileName(self, inputFileName):
        self._inFileName = inputFileName
        
    def setVerbosity(self, verbose):
        self._verbose = verbose
    
    def setAttributesFromCmdLine(self):
        description = "Insert banks in DB and launch PASTEC\n"
        description += "step 1 : insert profiles, repbaseNT and repbaseAA banks in DB\n"
        description += "step 2 : launch PASTEC\n"
        examples = "EXAMPLE TODO"
        parser = RepetOptionParser(description = description, version = "v1.0", epilog = examples)
        parser.add_option( "-C", "--config",        dest = "configFileName",        type = "string", action = "store", help = "configuration file (mandatory for step 1)", default = "")
        parser.add_option( "-D", "--decisionRules", dest = "decisionRulesFileName", type = "string", action = "store", help = "classification rules file name (e.g. TEclassifierRules.yml) [optional]", default = "")
        parser.add_option( "-P", "--project",       dest = "projectName",           type = "string", action = "store", help = "project name (if no configuration file or if different)", default = "")
        parser.add_option( "-S", "--step",          dest = "step",                  type = "string", action = "store", help = "step (default = 12, 1, 2)", default = "12")
        parser.add_option( "-i", "--input",         dest = "inputFileName",         type = "string", action = "store", help = "input file name (fasta)", default = "")
        parser.add_option( "-v", "--verbose",       dest = "verbose",               type = "int",    action = "store", help = "verbose (default = 0, 1)", default = 0)
        options = parser.parse_args()[0]
        self._setAttributesFromOptions(options)

    def _setAttributesFromOptions(self, options):
        self.setConfigFileName(options.configFileName)
        self.setDecisionRulesFileName(options.decisionRulesFileName)
        self.setProjectName(options.projectName)
        self.setStep(options.step)
        self.setInputFileName(options.inputFileName)
        self.setVerbosity(options.verbose)

    def _checkAttributes(self):
        if self._configFileName == "" and self._projectName == "":
            raise Exception("ERROR: No configuration file name or no project name specified in option!")
        if self._configFileName == "" and "1" in self._step:
            raise Exception("ERROR: No configuration file name specified in option!")
        if self._configFileName != "" and not FileUtils.isRessourceExists(self._configFileName):
            raise Exception("ERROR: Can't open configuration file : %s" % self._configFileName)
        if "2" in self._step:
            if self._inFileName == "":
                raise Exception("ERROR: No fasta file specified in option!")
            elif not FileUtils.isRessourceExists(self._inFileName):
                raise Exception("ERROR: Can't open fasta file : %s" % self._inFileName)

    def _createMapFile(self, profilesBank):
        tmpMapFile = "outputProfilesBank.map"
        profilesDB2Map = ProfilesDB2Map()
        profilesDB2Map.setInputFile(profilesBank)
        profilesDB2Map.setOutputFile(tmpMapFile)
        profilesDB2Map.run()
        return tmpMapFile
    
    def run(self):
        #TODO: refactoring + Check if table exists
        self._checkAttributes()
        if self._verbose > 0:
            sys.stdout.write("START LaunchPASTEC\n")
            sys.stdout.flush()
        
        config = None
        if self._configFileName != "":
            config = ConfigParser.ConfigParser()
            config.readfp(open(self._configFileName))
        if self._projectName == "":
            self._projectName = config.get("project", "project_name")
            
        if '1' in self._step:
            if self._verbose > 0:
                sys.stdout.write("step 1 : insert banks in DB\n")
                sys.stdout.flush()
            projectDir = config.get("project", "project_dir")
            sectionName = "detect_features"
            
            iDb = DbFactory.createInstance(self._configFileName)
            
            isHmmerUsed = config.get(sectionName, "TE_HMMER")
            if isHmmerUsed == "yes":
                profilesBank = "%s/%s" % (projectDir, config.get(sectionName, "TE_HMM_profiles"))
                tableName = "%s_ProfilesDB_map" % self._projectName
                if not iDb.doesTableExist(tableName):
                    if not FileUtils.isRessourceExists(profilesBank):
                        raise Exception("ERROR: can't find file '%s'\n" % profilesBank)
                    if self._verbose > 0:
                        msg = "- profiles bank: %s\n" % profilesBank
                        sys.stdout.write(msg)
                        sys.stdout.flush()
                    tmpMapFile = self._createMapFile(profilesBank)
                    iDb.createTable(tableName, "map", tmpMapFile)
                    os.remove(tmpMapFile)
                else:
                    msg = "- WARNING: %s already exists. Check it.\n" % tableName
                    sys.stdout.write(msg)
                    sys.stdout.flush()
                
            isBLRtxUsed = config.get(sectionName, "TE_BLRtx")
            isBLRnUsed = config.get(sectionName, "TE_BLRn")
            if isBLRtxUsed == "yes" or isBLRnUsed == "yes":
                repbaseNTBank = "%s/%s" % (projectDir, config.get(sectionName, "TE_nucl_bank"))
                #TODO: change "seq" to "length"
                tableName = "%s_RepbaseNT_seq" % self._projectName
                if not iDb.doesTableExist(tableName):
                    if not FileUtils.isRessourceExists(repbaseNTBank):
                        raise Exception("ERROR: can't find file '%s'\n" % repbaseNTBank)
                    if self._verbose > 0:
                        msg = "- repbaseNT bank: %s\n" % repbaseNTBank
                        sys.stdout.write(msg)
                        sys.stdout.flush()
                    iDb.createTable(tableName, "length", repbaseNTBank)
                else:
                    msg = "- WARNING: %s already exists. Check it.\n" % tableName
                    sys.stdout.write(msg)
                    sys.stdout.flush()
                
            isBLRxUsed = config.get(sectionName, "TE_BLRx")
            if isBLRxUsed == "yes":
                repbaseAABank = "%s/%s" % (projectDir, config.get(sectionName, "TE_prot_bank"))
                #TODO: change "seq" to "length"
                tableName = "%s_RepbaseAA_seq" % self._projectName
                if not iDb.doesTableExist(tableName):
                    if not FileUtils.isRessourceExists(repbaseAABank):
                        raise Exception("ERROR: can't find file '%s'\n" % repbaseAABank)
                    if self._verbose > 0:
                        msg = "- repbaseAA bank: %s\n" % repbaseAABank
                        sys.stdout.write(msg)
                        sys.stdout.flush()
                    iDb.createTable(tableName, "length", repbaseAABank)
                else:
                    msg = "- WARNING: %s already exists. Check it.\n" % tableName
                    sys.stdout.write(msg)
                    sys.stdout.flush()
            iDb.close()
                
        if '2' in self._step:
            outFileName = "%s.classif" % self._projectName
            if self._verbose > 0:
                msg = "step 2 : classification\n"
                msg += "- input file: %s\n" % self._inFileName
                msg += "- output file: %s\n" % outFileName
                sys.stdout.write(msg)
                sys.stdout.flush()
            
            iEnvironment = Environment(consensusFileName = self._inFileName, projectName = self._projectName, iConfig = config, decisionRulesFileName = self._decisionRulesFileName, verbose = self._verbose - 1)
            if config != None:
                iEnvironment.setAttributesFromConfig()
            iSuperAgent = SuperAgent(iEnvironment)
            iSuperAgent.launchAllAgents()
            iSuperAgent.live()
            
        if self._verbose > 0:
            msg = "END LaunchPASTEC\n"
            sys.stdout.write(msg)
            sys.stdout.flush()
    
if __name__ == '__main__':
    iLaunchPASTEC = LaunchPASTEC()
    iLaunchPASTEC.setAttributesFromCmdLine()
    iLaunchPASTEC.run()
