#!/usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from commons.core.LoggerFactory import LoggerFactory
from commons.core.utils.RepetOptionParser import RepetOptionParser

LOG_DEPTH = "repet.tools"

class ShortenConsensusHeader(object):
    
    def __init__(self, fastaFileName = "", outFileName = "", projectName = "", selfAlign = "blaster", clustering="", multAlign="", pyramidHeader=False, verbosity = 0):
        self._fastaFileName = fastaFileName
        self.setOutFileName(outFileName)
        self._projectName = projectName
        self._selfAlign = selfAlign
        self._clustering = clustering
        self._multAlign = multAlign
        self._pyramidHeader = pyramidHeader
        self._verbosity = verbosity
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self._verbosity)
        
    def setAttributesFromCmdLine(self):
        description = "Tool to shorten consensus headers from TEdenovo or SATannot.\n"
        parser = RepetOptionParser(description = description)
        parser.add_option("-i", "--fasta",          dest = "fastaFileName", action = "store",       type = "string", help = "input fasta file name [compulsory] [format: fasta]", default = "")
        parser.add_option("-o", "--out",            dest = "outFileName",   action = "store",       type = "string", help = "output file name [default=inFileName+'.shortH']", default = "")
        parser.add_option("-p", "--project",        dest = "projectName",   action = "store",       type = "string", help = "project name ", default = "")
        parser.add_option("-s", "--selfAlign",      dest = "selfAlign",     action = "store",       type = "string", help = "self-alignment program (Blaster/Pals) [default=blaster]", default = "blaster")
        parser.add_option("-c", "--clustering",     dest = "clustering",    action = "store",       type = "string", help = "clustering method (Grouper/Recon/Piler)[default=]", default = "")
        parser.add_option("-m", "--multAlign",      dest = "multAlign",     action = "store",       type = "string", help = "MSA program (Map/Mafft/Muscle/Clustal/Tcoffee/Prank/...)[default=]", default = "")
        parser.add_option("-H", "--pyramidHeader",  dest = "pyramidHeader", action = "store_true",                   help = "format the header with pyramid and piles informations (SATannot)[default=]", default = False)
        parser.add_option("-v", "--verbosity",      dest = "verbosity",     action = "store",       type = "int",    help = "verbosity [optional] [default: 1]", default = 1)
        options = parser.parse_args()[0]
        self._setAttributesFromOptions(options)
        
    def _setAttributesFromOptions(self, options):
        self._fastaFileName = options.fastaFileName
        self.setOutFileName(options.outFileName)
        self._projectName = options.projectName
        self._selfAlign = options.selfAlign
        self._clustering = options.clustering
        self._multAlign = options.multAlign
        self._pyramidHeader = options.pyramidHeader
        self.setVerbosity(options.verbosity)
        
    def setOutFileName(self, outFileName):
        if outFileName == "":
            self._outFileName = "%s.shortH" % self._fastaFileName
        else:
            self._outFileName = outFileName
        
    def setVerbosity(self, verbosity):
        self._verbosity = verbosity
        
    def _checkOptions(self):
        if self._fastaFileName == "":
            self._logAndRaise("ERROR: Missing input fasta file name")
            
    def _logAndRaise(self, errorMsg):
        self._log.error(errorMsg)
        raise Exception(errorMsg)
                    
    def run(self):
        LoggerFactory.setLevel(self._log, self._verbosity)
        self._checkOptions()
        self._log.info("START ShortenConsensusHeader")
        self._log.debug("Fasta file name: %s" % self._fastaFileName)
        
        outFile = open( self._outFileName, "w" )
        inFile = open( self._fastaFileName, "r" )
        lines = inFile.readlines()
    
        countSeq = 0
        for line in lines:
            if ">" in line:
                countSeq += 1
                data = line.split()
                cons_name = data[0].split("=")[1]
                cons_nbAlign = data[2].split("=")[1]
                if self._projectName != "":
                    if self._pyramidHeader:
                        cluster_id = data[3].split("=")[1]
                        pyramid_id = data[4].split("=")[1]
                        #TODO: option "header SATannot" => autre nom pour indiquer "PilerTA" ?
                        outFile.write( ">%s_%s_%s_%s_%s_%s_%s\n" % ( self._projectName, self._selfAlign, self._clustering, cluster_id, pyramid_id, self._multAlign, cons_nbAlign ) )
                    else:
                        cluster_id = cons_name.split("Cluster")[1]
                        cluster_id = cluster_id.split(".fa")[0]
                        outFile.write( ">%s_%s_%s_%s_%s_%s\n" % (self._projectName, self._selfAlign, self._clustering, cluster_id, self._multAlign, cons_nbAlign ) )
                else:
                    outFile.write( ">%s_nbAlign%s\n" % ( cons_name, cons_nbAlign ) )
            else:
                seq = line.split("\n")
                outFile.write(seq[0]+"\n")
    
        inFile.close()
        outFile.close()
    
        self._log.debug("%i consensus in %s" % (countSeq, self._fastaFileName))
        self._log.info("END ShortenConsensusHeader")

if __name__ == "__main__":
    iShortenConsensusHeader = ShortenConsensusHeader()
    iShortenConsensusHeader.setAttributesFromCmdLine()
    iShortenConsensusHeader.run()        